;*---------------------------------------------------------------------*/
;*   A pratical implementation for the Scheme programming language     */
;*                                                                     */
;*                                    ,--^,                            */
;*                              _ ___/ /|/                             */
;*                          ,;'( )__, ) '                              */
;*                         ;;  //   L__.                               */
;*                         '   \\   /  '                               */
;*                              ^   ^                                  */
;*                                                                     */
;*   Copyright (c) 1992-1999 Manuel Serrano                            */
;*                                                                     */
;*     Bug descriptions, use reports, comments or suggestions are      */
;*     welcome. Send them to                                           */
;*       bigloo-request@kaolin.unice.fr                                */
;*       http://kaolin.unice.fr/bigloo                                 */
;*                                                                     */
;*   This program is free software; you can redistribute it            */
;*   and/or modify it under the terms of the GNU General Public        */
;*   License as published by the Free Software Foundation; either      */
;*   version 2 of the License, or (at your option) any later version.  */
;*                                                                     */
;*   This program is distributed in the hope that it will be useful,   */
;*   but WITHOUT ANY WARRANTY; without even the implied warranty of    */
;*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     */
;*   GNU General Public License for more details.                      */
;*                                                                     */
;*   You should have received a copy of the GNU General Public         */
;*   License along with this program; if not, write to the Free        */
;*   Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,   */
;*   MA 02111-1307, USA.                                               */
;*---------------------------------------------------------------------*/
;*---------------------------------------------------------------------*/
;*    serrano/prgm/project/bigloo/runtime/Eval/expd-let.scm            */
;*                                                                     */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Mon Jan  4 17:10:13 1993                          */
;*    Last change :  Fri Sep 18 17:55:49 1998 (serrano)                */
;*                                                                     */
;*    Les expanseurs des formes `let's                                 */
;*---------------------------------------------------------------------*/

;*---------------------------------------------------------------------*/
;*    Le module                                                        */
;*---------------------------------------------------------------------*/
(module __expander_let
   
   (import  (__error                   "Llib/error.scm")
	    (__bigloo                  "Llib/bigloo.scm")
	    (__tvector                 "Llib/tvector.scm")
	    (__structure               "Llib/struct.scm")
	    (__tvector                 "Llib/tvector.scm")
	    (__bexit                   "Llib/bexit.scm")
	    
	    (__r4_numbers_6_5          "Ieee/number.scm")
	    (__r4_numbers_6_5_fixnum   "Ieee/fixnum.scm")
	    (__r4_numbers_6_5_flonum   "Ieee/flonum.scm")
	    (__r4_characters_6_6       "Ieee/char.scm")
	    (__r4_equivalence_6_2      "Ieee/equiv.scm")
	    (__r4_booleans_6_1         "Ieee/boolean.scm")
	    (__r4_symbols_6_4          "Ieee/symbol.scm")
	    (__r4_strings_6_7          "Ieee/string.scm")
	    (__r4_pairs_and_lists_6_3  "Ieee/pair-list.scm")
	    (__r4_input_6_10_2         "Ieee/input.scm")
	    (__r4_control_features_6_9 "Ieee/control.scm")
	    (__r4_vectors_6_8          "Ieee/vector.scm")
	    (__r4_ports_6_10_1         "Ieee/port.scm")
	    (__r4_output_6_10_3        "Ieee/output.scm")
	    
	    (__progn                   "Eval/progn.scm"))
   
   (use     (__type                    "Llib/type.scm")
	    (__evenv                   "Eval/evenv.scm"))
   
   (export  (expand-eval-let    <expression> <expander>)
	    (expand-eval-let*   <expression> <expander>)
	    (expand-eval-letrec <expression> <expander>)
	    (expand-eval-labels <expression> <expander>)))

;*---------------------------------------------------------------------*/
;*    expand-eval-let ...                                              */
;*---------------------------------------------------------------------*/
(define (expand-eval-let x e)
   (let ((res (match-case x
		 ((?- () . (and ?body (not ())))
		  (e `((lambda () ,(normalize-progn body))) e))
		 ((?- (and (? symbol?) ?loop) ?bindings . (and ?body (not ())))
		  (if (not (or (null? bindings)
			       (pair? bindings)))
		      (error "let" "Illegal `let' form" x)
		      (e `(letrec ((,loop (lambda
						,(map
						  (lambda (b)
						     (if (pair? b)
							 (car b)
							 (error
							  "expand-let"
							  "Illegal `let' form"
							  x)))
						      bindings)
					     ,(normalize-progn body))))
			     (,loop ,@(map (lambda (b)
					      (cadr b))
					   bindings)))
			 e)))
		 ((?- (and (? pair?) ?bindings) . (and ?body (not ())))
		  (let ((body      (normalize-progn body))
			(vars.vals (let loop ((bindings bindings)
					      (vars     '())
					      (vals     '()))
				      (cond
					 ((null? bindings)
					  (cons vars vals))
					 ((not (pair? (car bindings)))
					  (loop (cdr bindings)
						(cons (car bindings) vars)
						(cons '(unspecified) vals)))
					 (else
					  (loop (cdr bindings)
						(cons (car (car bindings))
						      vars)
						(cons (normalize-progn
						       (cdr (car bindings)))
						      vals)))))))
		     (e `((lambda ,(car vars.vals) ,body)
			  ,@(cdr vars.vals)) e)))
		 (else
		  (error "expand-let" "Illegal `let' form" x)))))
      (replace! x res)))
	   
;*---------------------------------------------------------------------*/
;*    expand-eval-let* ...                                             */
;*---------------------------------------------------------------------*/
(define (expand-eval-let* x e)
   (let ((res (match-case x
		 ((?- () . (and ?body (not ())))
		  (e `((lambda () ,(normalize-progn body))) e))
		 ((?- ?bindings . (and ?body (not ())))
		  (e `(let (,(car bindings))
			 (let* ,(cdr bindings)
			    ,(normalize-progn body))) e))
		 (else
		  (error "expand-let*" "Illegal form" x)))))
      (replace! x res)))
   
;*---------------------------------------------------------------------*/
;*    expand-eval-letrec ...                                           */
;*---------------------------------------------------------------------*/
(define (expand-eval-letrec x e)
   (let ((res (match-case x
		 ((?- () . (and ?body (not ())))
		  (e `((lambda () ,(normalize-progn body))) e))
		 ((?- (and (? pair?) ?bindings) . (and ?body (not ())))
		  (let ((body      (normalize-progn body))
			(vars.vals (let loop ((bindings bindings)
					      (vars     '())
					      (vals     '()))
				      (cond
					 ((null? bindings)
					  (cons vars vals))
					 ((not (pair? (car bindings)))
					  (error "letrec" "Illegal form" x))
					 (else
					  (loop (cdr bindings)
						(cons (car (car bindings))
						      vars)
						(cons (normalize-progn
						       (cdr (car bindings)))
						      vals)))))))
		     (e `((lambda ,(car vars.vals)
			     ,(normalize-progn
			       (let loop ((vars (car vars.vals))
					  (vals (cdr vars.vals))
					  (res  (list body)))
				  (if (null? vars)
				      res
				      (loop (cdr vars)
					    (cdr vals)
					    (cons `(set! ,(car vars)
							 ,(car vals))
						  res))))))
			  ,@(map (lambda (v) '(unspecified)) bindings)) e)))
		 (else
		  (error "letrec" "Illegal form" x)))))
      (replace! x res)))

;*---------------------------------------------------------------------*/
;*    expand-eval-labels ...                                           */
;*---------------------------------------------------------------------*/
(define (expand-eval-labels x e)
   (let ((res (match-case x
		 ((?- () . (and ?body (not ())))
		  (e `((lambda () ,(normalize-progn body))) e))
		 ((?- ?bindings . (and ?body (not ())))
		  (let ((new (let loop ((bindings bindings))
				(cond
				   ((null? bindings)
				    '())
				   ((not (pair? bindings))
				    (error "expand-labels" "Illegal form" x))
				   (else
				    (match-case (car bindings)
				       ((?name ?args . ?lbody)
					(cons `(,name (lambda ,args ,@lbody))
					      (loop (cdr bindings))))
				       (else
					(error "expand-labels"
					       "Illegal form"
					       x))))))))
		     (e `(letrec ,new ,@body) e)))
		 (else
		  (error "expand-labels" "Illegal form" x)))))
      (replace! x res)))
